#!/usr/bin/env python
# coding: utf-8

# # Fibonnaci
# 
# In this notebook we will show a few different ways of compiling the fibonnaci functon with LLVM in Python.
# 
# This is inspired by [Siu's work](https://github.com/sklam/etude-okvmap/blob/1e85c5ea7bb1f092ed251159e7046d1348f3fa8b/example_minilang.py#L285-L300) as an attempt to explore how metadsl could be useful as a way to build higher level expression systems on top of LLVM in Python.

# ## Pure Python
# 
# First, lets create a pure python version

# In[1]:


N = 1000


# In[2]:


def fib_more(n, a, b):
    if n > 1:
        return fib_more(n - 1, b, a + b)
    if b == 1:
        return b
    return a

def fib(n):
    return fib_more(n, 0, 1)


# In[3]:


# %timeit fib(N)


# ## `llvmlite`
# 
# First, lets start with the low level `llvmlite` library to build up the llvm directly:

# In[6]:


from llvmlite import ir

def create_mod():
    mod = ir.Module(name="fib")

    int_type = ir.IntType(32)
    zero = ir.Constant(int_type, 0)
    one = ir.Constant(int_type, 1)

    fib_more_fn = ir.Function(
        mod,
        ir.FunctionType(int_type, [int_type, int_type, int_type]),
        name='fib_more',
    )
    
    fib_fn = ir.Function(
        mod,
        ir.FunctionType(int_type, [int_type]),
        name='fib'
    )
    
    fib_fn.calling_convention = 'fastcc'
    fib_more_fn.calling_convention = 'fastcc'


    fib_n, = fib_fn.args
    fib_more_n, fib_more_a, fib_more_b = fib_more_fn.args


    fib_entry = fib_fn.append_basic_block('entry')
    fib_more_entry = fib_more_fn.append_basic_block('entry')
    fib_pred_cont = fib_more_fn.append_basic_block('pred_cont')
    fib_not_pred_cont = fib_more_fn.append_basic_block('not_pred_cont')
    fib_n_eq_one = fib_more_fn.append_basic_block('n_eq_one')
    fib_n_neq_one = fib_more_fn.append_basic_block('n_neq_one')

    fib_entry_builder = ir.IRBuilder(fib_entry)
    fib_more_entry_builder = ir.IRBuilder(fib_more_entry)
    fib_pred_cont_builder = ir.IRBuilder(fib_pred_cont)
    fib_not_pred_cont_builder = ir.IRBuilder(fib_not_pred_cont)
    fib_n_eq_one_builder = ir.IRBuilder(fib_n_eq_one)
    fib_n_neq_one_builder = ir.IRBuilder(fib_n_neq_one)


    fib_entry_builder.ret(
        fib_entry_builder.call(
            fib_more_fn,
            (fib_n, zero, one)
        )
    )


    pred_cont = fib_more_entry_builder.icmp_signed(">", fib_more_n, one)


    fib_more_entry_builder.cbranch(pred_cont, fib_pred_cont, fib_not_pred_cont)
    
    minus1 = fib_pred_cont_builder.sub(fib_more_n, one)
    ab = fib_pred_cont_builder.add(fib_more_a, fib_more_b)
    added = fib_pred_cont_builder.call(fib_more_fn, (minus1, fib_more_b, ab))
    fib_pred_cont_builder.ret(added)

    n_eq_1 = fib_not_pred_cont_builder.icmp_signed("==", fib_more_n, one)
    fib_not_pred_cont_builder.cbranch(n_eq_1, fib_n_eq_one, fib_n_neq_one)

    fib_n_eq_one_builder.ret(fib_more_b)

    fib_n_neq_one_builder.ret(fib_more_a)

    return mod

llvm_ir = (str(create_mod()))
# print(llvm_ir)


# In[5]:


from ctypes import CFUNCTYPE, c_int
from llvmlite import binding as llvm

llvm.initialize()
llvm.initialize_native_target()
llvm.initialize_native_asmprinter()

def make_c_wrapper(fn_callee):
    mod = fn_callee.module
    fnty = fn_callee.function_type
    fn = ir.Function(mod, fnty, name='entry_' + fn_callee.name)
    builder = ir.IRBuilder(fn.append_basic_block())
    builder.ret(builder.call(fn_callee, fn.args))



def execute(llvm_ir):
    """
    Compile the LLVM IR string with the given engine.
    The compiled module object is returned.
    """
    llmod = llvm.parse_assembly(str(llvm_ir))
    llmod.verify()
    
#     print('optimized'.center(80, '-'))
    pmb = llvm.create_pass_manager_builder()
    pmb.opt_level = 1
    pm = llvm.create_module_pass_manager()
    pmb.populate(pm)
    pm.run(llmod)
#     print(llmod)

    target_machine = llvm.Target.from_default_triple().create_target_machine()

    ee = llvm.create_mcjit_compiler(llmod, target_machine)
    ee.finalize_object()
    cfptr = ee.get_function_address("entry_fib")

    cfunc = CFUNCTYPE(c_int, c_int)(cfptr)
    return ee, cfunc

    # Get CFG
#         ll_fib_more = llmod.get_function('fib_more')
#         cfg = llvm.get_function_cfg(ll_fib_more)
#         llvm.view_dot_graph(cfg, view=True)

mod = create_mod()
make_c_wrapper(mod.get_global('fib'))
_, f = execute(mod)


# In[6]:


# %timeit f(N)


# We see this is two orders of magnitude faster than the Python code, partially because we are closer to the metal and partially because LLVM turns this tail recursion into a loop.

# ## `metadsl` with `llvmlite` wrapper

# In[1]:


import metadsl as m
import metadsl_core as mc
import metadsl_llvm as ml
import metadsl_visualize


# In[17]:



def create_metadsl_fn():
    # Constants
    int_type = ml.Type.create_int(32)
    zero = ml.Value.constant(int_type, 0)
    one = ml.Value.constant(int_type, 1)

    # Module reference
    mod_ref = ml.ModuleReference.create("fib")

    # Module builder
    mod_builder = ml.ModuleBuilder.create(mod_ref)

    # Function references

    mod_builder, fib_more_ref = ml.FunctionReference.create(
        mod_builder,
        ml.FunctionType.create(int_type, int_type, int_type, int_type),
        "fib_more",
        "fastcc",
    ).spread

    mod_builder, fib_ref = ml.FunctionReference.create(
        mod_builder, ml.FunctionType.create(int_type, int_type), "fib", "fastcc"
    ).spread

    # Function builders
    fib_more_builder = ml.FunctionBuilder.create(fib_more_ref)
    fib_builder = ml.FunctionBuilder.create(fib_ref)

    # Function args
    fib_n = fib_more_builder.arguments[mc.Integer.from_int(0)]

    fib_more_n = fib_builder.arguments[mc.Integer.from_int(0)]
    fib_more_a = fib_builder.arguments[mc.Integer.from_int(1)]
    fib_more_b = fib_builder.arguments[mc.Integer.from_int(2)]

    # Block references
    fib_more_builder, fib_more_entry_ref = ml.BlockReference.create(
        "entry", fib_more_builder
    ).spread
    fib_more_builder, fib_more_pred_cont_ref = ml.BlockReference.create(
        "pred_cont", fib_more_builder
    ).spread
    fib_more_builder, fib_more_not_pred_cont_ref = ml.BlockReference.create(
        "not_pred_cont", fib_more_builder
    ).spread
    fib_more_builder, fib_more_n_eq_one_ref = ml.BlockReference.create(
        "n_eq_one", fib_more_builder
    ).spread
    fib_more_builder, fib_more_n_neq_one_ref = ml.BlockReference.create(
        "n_neq_one", fib_more_builder
    ).spread

    fib_builder, fib_entry_ref = ml.BlockReference.create("entry", fib_builder).spread

    # Block builders

    fib_more_entry_builder = ml.Builder.create(fib_more_entry_ref)
    fib_more_pred_cont_builder = ml.Builder.create(fib_more_pred_cont_ref)
    fib_more_not_pred_cont_builder = ml.Builder.create(fib_more_not_pred_cont_ref)
    fib_more_n_eq_one_builder = ml.Builder.create(fib_more_n_eq_one_ref)
    fib_more_n_neq_one_builder = ml.Builder.create(fib_more_n_neq_one_ref)

    fib_entry_builder = ml.BlockBuilder.create(fib_entry_ref)

    # Values

    fib_entry_builder, fib_entry_value = fib_entry_builder.call(
        fib_more_ref, mc.Vec.create(fib_n, zero, one)
    ).spread
    fib_entry_builder = fib_entry_builder.ret(fib_entry_value)

    fib_more_entry_builder, pred_cont = fib_more_entry_builder.icmp_signed(
        ">", fib_more_n, one
    ).spread
    fib_more_entry_builder = fib_more_entry_builder.cbranch(
        pred_cont, fib_more_pred_cont_ref, fib_more_not_pred_cont_ref
    )

    fib_more_pred_cont_builder, minus1 = fib_more_pred_cont_builder.sub(
        fib_more_n, one
    ).spread
    fib_more_pred_cont_builder, ab = fib_more_pred_cont_builder.add(
        fib_more_a, fib_more_b
    )
    fib_more_pred_cont_builder, added = fib_more_pred_cont_builder.call(
        fib_more_ref, mc.Vec.create(minus1, fib_more_b, ab)
    ).spread
    fib_more_pred_cont_builder = fib_more_pred_cont_builder.ret(added)

    fib_more_not_pred_cont_builder, n_eq_1 = fib_more_not_pred_cont_builder.icmp_signed(
        "==", fib_more_n, one
    )
    fib_more_not_pred_cont_builder = fib_more_not_pred_cont_builder.cbranch(
        n_eq_1, fib_more_n_eq_one_ref, fib_more_n_neq_one_ref
    )

    fib_more_n_eq_one_builder = fib_more_n_eq_one_builder.ret(fib_more_b)

    fib_more_n_neq_one_builder = fib_more_n_neq_one_builder.ret(fib_more_a)

    # Blocks
    fib_more_entry_block = ml.Block.create(fib_more_entry_ref, fib_more_entry_builder)
    fib_more_pred_cont_block = ml.Block.create(
        fib_more_pred_cont_ref, fib_more_pred_cont_builder
    )
    fib_more_not_pred_cont_block = ml.Block.create(
        fib_more_not_pred_cont_ref, fib_more_not_pred_cont_builder
    )
    fib_more_n_eq_one_block = ml.Block.create(
        fib_more_n_eq_one_ref, fib_more_n_eq_one_builder
    )
    fib_more_n_neq_one_block = ml.Block.create(
        fib_more_n_neq_one_ref, fib_more_n_neq_one_builder
    )

    fib_entry_block = ml.Block.create(fib_entry_ref, fib_entry_builder)

    # Functions
    fib_more_fn = ml.Function.create(
        fib_more_ref,
        mc.Vec.create(
            fib_more_entry_block,
            fib_more_pred_cont_block,
            fib_more_not_pred_cont_block,
            fib_more_n_eq_one_block,
            fib_more_n_neq_one_block,
        ),
    )

    fib_fn = ml.Function.create(fib_ref, mc.Vec.create(fib_entry_block))

    # Module

    module = ml.Module.create(mod_ref, mc.Vec.create(fib_fn, fib_more_fn))

    c_int = ml.CType.c_int()
    c_func_type = ml.CFunctionType.create(c_int, c_int)

    return ml.compile_function(module, mod_builder, fib_ref, c_func_type)


# In[18]:


metadsl_fn = m.execute(create_metadsl_fn())


# In[ ]:







from __future__ import annotations

import inspect
import typing
import copy
import typing_inspect
import collections.abc
import dataclasses
import functools
import types
from .dict_tools import *

__all__ = [
    "get_type",
    "get_arg_hints",
    "GenericCheck",
    "infer",
    "TypeVarMapping",
    "OfType",
    "replace_typevars",
    "match_functions",
    "match_values",
    "BoundInfer",
    "TypeVarScope",
    "infer_return_type",
    "ExpandedType",
    "replace_fn_typevars",
    "merge_typevars",
    "match_types",
    "get_origin_type",
    "get_fn_typevars",
]

T = typing.TypeVar("T")
U = typing.TypeVar("U")


class GenericCheckType(type):
    @functools.lru_cache()
    def __subclasscheck__(cls, sub):
        """
        Modified from https://github.com/python/cpython/blob/aa73841a8fdded4a462d045d1eb03899cbeecd65/Lib/typing.py#L707-L717
        """
        sub = getattr(sub, "__origin__", sub)
        if hasattr(cls, "__origin__"):
            return issubclass(sub, cls)
        if sub == typing.Any:
            return False

        # Needed when checking if T or Union is instance of Expression
        if isinstance(sub, typing.TypeVar) or sub == typing.Union:  # type: ignore
            return False
        return super().__subclasscheck__(sub)


class GenericCheck(metaclass=GenericCheckType):
    """
    Subclass this to support isinstance and issubclass checks with generic classes.
    """

    pass


class OfType(GenericCheck, typing.Generic[T]):
    """
    OfType[T] should be considered a subclass of T even though it is not.
    """

    pass


class ExpandedType(GenericCheck, typing.Generic[T]):
    """
    ExpandedType should be thought of as being expanded when passed into a function,
    so that `fn(ExpandedType[int]())` will be thought of as `fn(*xs)` where xs is an iterable of `int`.
    """

    pass


original_generic_getattr = typing._GenericAlias.__getattr__  # type: ignore


def generic_getattr(self, attr):
    """
    Allows classmethods to get generic types
    by checking if we are getting a descriptor type
    and if we are, we pass in the generic type as the class
    instead of the origin type.

    Modified from
    https://github.com/python/cpython/blob/aa73841a8fdded4a462d045d1eb03899cbeecd65/Lib/typing.py#L694-L699
    """
    if "__origin__" in self.__dict__ and attr not in (
        "__wrapped__",
        "__union_params__",
    ):
        # If the attribute is a descriptor, pass in the generic class
        try:
            property = self.__origin__.__getattribute__(self.__origin__, attr)
        except Exception:
            return

        if hasattr(property, "__get__"):
            return property.__get__(None, self)
        # Otherwise, just resolve it normally
        return getattr(self.__origin__, attr)
    raise AttributeError(attr)


typing._GenericAlias.__getattr__ = generic_getattr  # type: ignore


# Allow isinstance and issubclass calls on generic types
def generic_subclasscheck(self, cls):
    """
    Modified from https://github.com/python/cpython/blob/aa73841a8fdded4a462d045d1eb03899cbeecd65/Lib/typing.py#L707-L717
    """
    cls = getattr(cls, "__origin__", cls)
    return issubclass(self.__origin__, cls)


# Allow isinstance and issubclass calls on special forms like union
def special_form_subclasscheck(self, cls):
    if self == typing.Any:
        return True
    if self == cls:
        return True
    raise TypeError


typing._GenericAlias.__subclasscheck__ = generic_subclasscheck  # type: ignore
typing._SpecialForm.__subclasscheck__ = special_form_subclasscheck  # type: ignore


def get_origin(t: typing.Type) -> typing.Type:
    origin = typing_inspect.get_origin(t)
    # Need workaround for sequences
    # https://github.com/ilevkivskyi/typing_inspect/issues/36
    if origin == collections.abc.Sequence:
        return typing.Sequence

    if origin == tuple:
        return typing.Tuple  # type: ignore

    return origin


def get_function_type(fn: typing.Callable) -> typing.Type[typing.Callable]:
    """
    Gets the type of a function:

    Only supports positional args currently

    >>> get_function_type(lambda i: i)
    typing.Callable[[typing.Any], typing.Any]

    >>> def fn(a: int, b: str) -> float: ...
    >>> get_function_type(fn)
    typing.Callable[[int, str], float]

    >>> def no_return_type(a: int, b: str): ...
    >>> get_function_type(no_return_type)
    typing.Callable[[int, str], typing.Any]

    >>> def no_arg_type(a, b: str): ...
    >>> get_function_type(no_arg_type)
    typing.Callable[[typing.Any, str], typing.Any]
    """
    signature = inspect_signature(fn)
    type_hints = typing_get_type_hints(fn)
    arg_hints: typing.List[typing.Type] = []

    for arg_name, p in signature.parameters.items():
        if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            arg_hints.append(
                type_hints.get(arg_name, typing.cast(typing.Type, typing.Any))
            )
        else:
            raise NotImplementedError(f"Does not support getting type of {signature}")
    return typing.Callable[arg_hints, type_hints.get("return", typing.Any)]


def get_function_replace_type(f: FunctionReplaceTyping) -> typing.Type[typing.Callable]:
    return replace_typevars(f.typevars, get_function_type(f.fn))


def get_bound_infer_type(b: BoundInfer) -> typing.Type[typing.Callable]:
    """
    Returns a typing.Callable type that corresponds to the type of the bound infer.

    TODO: This logic is a combination of `get_function_type` and `infer_return_type`.
    We should eventually merge all of this into a consistant API so we don't have to duplicate this code.
    """
    hints = copy.copy(typing_get_type_hints(b.fn))
    signature = inspect_signature(b.fn)

    # We want to get the original type hints for the first arg,
    # and match those against the first arg in the bound, so we get a typevar mapping
    first_arg_name = next(iter(signature.parameters.keys()))
    typevars: TypeVarMapping
    # Whether to skip the first arg of the param of the signature when computing the signature
    skip_first_param: bool
    owner_origin = get_origin_type(b.owner)
    if b.is_classmethod:
        # If we called this as a class method
        typevars = match_type(typing.Type[owner_origin], b.owner)
        skip_first_param = True
    else:
        # we are calling an instance method on the class and passing the instance as the first arg
        typevars = match_type(typing.Type[owner_origin], b.owner)
        skip_first_param = False
        hints[first_arg_name] = owner_origin

    # Then we want to replace all the typevar hints, with what we now know from the first arg
    arg_hints: typing.List[typing.Type] = []

    for arg_name, p in signature.parameters.items():
        if skip_first_param:
            skip_first_param = False
            continue
        if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            arg_hints.append(
                replace_typevars(
                    typevars, hints.get(arg_name, typing.cast(typing.Type, typing.Any))
                )
            )
        else:
            raise NotImplementedError(f"Does not support getting type of {signature}")
    return typing.Callable[
        arg_hints,
        replace_typevars(
            typevars, hints.get("return", typing.cast(typing.Type, typing.Any))
        ),
    ]


def get_type(v: T) -> typing.Type[T]:
    """
    Returns the type of the value with generic arguments preserved.
    """
    if isinstance(v, functools.partial):  # type: ignore
        inner_type = get_type(v.func)  # type: ignore
        if v.keywords:  # type: ignore
            raise TypeError
        inner_args, inner_return = typing_inspect.get_args(inner_type)
        mapping: TypeVarMapping = merge_typevars(
            *(
                match_type(arg_type, arg)
                for (arg_type, arg) in zip(inner_args, v.args)  # type: ignore
            )
        )
        rest_arg_types = inner_args[len(v.args) :]  # type: ignore
        return typing.Callable[
            [replace_typevars(mapping, arg) for arg in rest_arg_types],
            replace_typevars(mapping, inner_return),
        ]

    if isinstance(v, Infer):
        return get_function_type(v.fn)  # type: ignore
    if isinstance(v, BoundInfer):
        return get_bound_infer_type(v)  # type: ignore
    if isinstance(v, FunctionReplaceTyping):
        return get_function_replace_type(v)  # type: ignore

    tp = typing_inspect.get_generic_type(v)
    # Special case, only support homogoneous tuple that are inferred to iterables
    if tp == tuple:
        if v:
            return typing.Sequence[get_type(v[0])]  # type: ignore
        return typing.Sequence  # type: ignore
    # Special case, also support function types.
    if tp == types.FunctionType:
        return get_function_type(v)  # type: ignore

    return tp


TypeVarMapping = typing.Mapping[typing.TypeVar, typing.Type]  # type: ignore


def match_values(hint_value: T, value: T) -> TypeVarMapping:
    return match_type(get_type(hint_value), value)


def match_type(hint: typing.Type[T], value: T) -> TypeVarMapping:
    if typing_inspect.get_origin(hint) == type:
        (inner_hint,) = typing_inspect.get_args(hint)
        return match_types(inner_hint, typing.cast(typing.Type, value))
    return match_types(hint, get_type(value))


def merge_typevars(*typevars: TypeVarMapping) -> TypeVarMapping:
    """
    Merges typevar mappings. If there is a duplicate key, either the values should be the
    same or one should have `typing.Any` type, or one should be a typevar itself. If it is a typevar,
    that typevar is also set to the other's value
    """
    merged: TypeVarMapping = {}
    typevars_: typing.List[TypeVarMapping] = list(typevars)
    while typevars_:
        tvs: TypeVarMapping = typevars_.pop()
        for tv, tp in tvs.items():  # type: ignore
            if tv not in merged:
                merged[tv] = tp  # type: ignore
                continue
            prev_tp = merged[tv]
            if prev_tp == typing.Any:
                merged[tv] = tp  # type: ignore
            elif prev_tp == tp or tp == typing.Any:
                pass
            elif typing_inspect.is_typevar(prev_tp):
                merged[prev_tp] = merged[tv] = tp  # type: ignore
            elif typing_inspect.is_typevar(tp):
                merged[tp] = prev_tp  # type: ignore
            else:
                # Try merging them and choosing replacing the hint with the
                # merged values
                try:
                    merged[tv] = replace_typevars(  # type: ignore
                        match_types(prev_tp, tp), prev_tp
                    )
                    continue
                except TypeError:
                    pass
                try:
                    merged[tv] = replace_typevars(  # type: ignore
                        match_types(tp, prev_tp), tp
                    )
                    continue
                except TypeError:
                    pass
                raise TypeError(f"Cannot merge {prev_tp} and {tp} for type var {tv}")
    return merged


def get_inner_types(t: typing.Type) -> typing.Iterable[typing.Type]:
    """
    Returns the inner types for a type.

    Like `typing_inspect.get_args` but special cases callable, so it returns
    the return type, then all the arg types.
    """
    if t == typing.Callable:  # type: ignore
        return []
    if typing_inspect.get_origin(t) == collections.abc.Callable:
        arg_types, return_type = typing_inspect.get_args(t)
        return [return_type] + arg_types
    return typing_inspect.get_args(t)


def get_all_typevars(t: typing.Type) -> typing.Iterable[typing.TypeVar]:  # type: ignore
    for t in get_inner_types(t):
        if isinstance(t, typing.TypeVar):  # type: ignore
            yield t
        else:
            yield from get_all_typevars(t)


def match_types(hint: typing.Type, t: typing.Type) -> TypeVarMapping:
    """
    Matches a type hint with a type, return a mapping of any type vars to their values.
    """
    if hint == t:
        return {}

    # If it is an instance of OfType[Type[T]], then we should consider it as T
    if isinstance(t, OfType):
        (of_type,) = typing_inspect.get_args(get_type(t))
        assert issubclass(of_type, typing.Type)
        (t,) = typing_inspect.get_args(of_type)
        return match_types(hint, t)

    # If the type is an OfType[T] then we should really just consider it as T
    if issubclass(t, OfType) and not issubclass(hint, OfType):
        (t,) = typing_inspect.get_args(t)
        return match_types(hint, t)
    if issubclass(hint, OfType) and not issubclass(t, OfType):
        (hint,) = typing_inspect.get_args(hint)
        return match_types(hint, t)

    # Matching an expanded type is like matching just whatever it represents
    if issubclass(t, ExpandedType):
        (t,) = typing_inspect.get_args(t)

    if typing_inspect.is_typevar(hint):
        return {hint: t}

    # This happens with match rule on conversion, like when the value is TypeVar
    if typing_inspect.is_typevar(t):
        return {}

    # if both are generic sequences, verify they are the same and have the same contents
    if (
        typing_inspect.is_generic_type(hint)
        and typing_inspect.is_generic_type(t)
        and typing_inspect.get_origin(hint) == collections.abc.Sequence
        and typing_inspect.get_origin(t) == collections.abc.Sequence
    ):
        t_inner = typing_inspect.get_args(t)[0]

        # If t's inner arg is just the default one for seuqnce, it hasn't be initialized so assume
        # it was an empty tuple that created it and just return a match
        if t_inner == typing_inspect.get_args(typing.Sequence)[0]:
            return {}
        return match_types(typing_inspect.get_args(hint)[0], t_inner)

    if typing_inspect.is_union_type(hint):
        # If this is a union, iterate through and use the first that is a subclass
        for inner_type in typing_inspect.get_args(hint):
            if issubclass(t, inner_type):
                hint = inner_type
                break
        else:
            raise TypeError(f"Cannot match concrete type {t} with hint {hint}")

    if not issubclass(t, hint):
        raise TypeError(f"Cannot match concrete type {t} with hint {hint}")
    return merge_typevars(
        *(
            match_types(inner_hint, inner_t)
            for inner_hint, inner_t in zip(get_inner_types(hint), get_inner_types(t))
        )
    )


def get_origin_type(t: typing.Type) -> typing.Type:
    """
    Takes in a type, and if it is a generic type, returns the original type hint.

    typing.List[int] -> typing.List[T]
    """
    t = get_origin(t) or t

    params = typing_inspect.get_parameters(t)
    if params:
        return t[params]
    return t


T_type = typing.TypeVar("T_type", bound=typing.Type)


def replace_typevars(typevars: TypeVarMapping, hint: T_type) -> T_type:
    """
    Replaces type vars in a type hint with other types.
    """
    if typing_inspect.is_typevar(hint):
        return typing.cast(T_type, typevars.get(hint, hint))

    # Special case empty callable, which raisees error on getting args
    if hint == typing.Callable:  # type: ignore
        return hint
    if typing_inspect.get_origin(hint) == collections.abc.Callable:
        arg_types, return_type = typing_inspect.get_args(hint)
        return typing.Callable[
            [replace_typevars(typevars, a) for a in arg_types],
            replace_typevars(typevars, return_type),
        ]

    args = typing_inspect.get_args(hint)
    if not args:
        return hint
    replaced_args = tuple(replace_typevars(typevars, arg) for arg in args)
    return get_origin(hint)[replaced_args]


@functools.lru_cache()
def inspect_signature(fn: typing.Callable) -> inspect.Signature:
    return inspect.signature(fn)


@functools.lru_cache()
def typing_get_type_hints(fn: typing.Callable) -> typing.Dict[str, typing.Type]:
    return typing.get_type_hints(fn)


def get_arg_hints(fn: typing.Callable) -> typing.List[typing.Type]:
    signature = inspect_signature(fn)
    hints = typing_get_type_hints(fn)
    return [hints[param] for param in signature.parameters.keys()]


def infer_return_type(
    fn: typing.Callable[..., T],
    owner: typing.Optional[typing.Type],
    is_classmethod: bool,
    args: typing.Tuple[object, ...],
    kwargs: typing.Mapping[str, object],
) -> typing.Tuple[
    typing.Tuple[object, ...],
    typing.Mapping[str, object],
    typing.Type[T],
    TypeVarMapping,
]:
    hints = copy.copy(typing_get_type_hints(fn))
    signature = inspect_signature(fn)

    mappings: typing.List[TypeVarMapping] = []
    # This case is triggered if we got here from a __get__ call
    # in a descriptor
    if owner:
        first_arg_name = next(iter(signature.parameters.keys()))
        first_arg_type: typing.Type
        owner_origin = get_origin_type(owner)
        if is_classmethod:
            # If we called this as a class method, add the owner to
            # the args add the inferred type to the hints.
            args = (owner,) + args  # type: ignore
            first_arg_type = typing.Type[owner_origin]  # type: ignore
        else:
            # If the owner had type parameters set, we should use those to start computing variables
            # i.e. Class[int].__add__
            mappings.append(match_types(owner_origin, owner))
            first_arg_type = owner_origin  # type: ignore

        if first_arg_name not in hints:
            hints[first_arg_name] = first_arg_type  # type: ignore

    bound = signature.bind(*args, **kwargs)
    bound.apply_defaults()

    # We need to edit the arguments to pop off the variable one
    arguments = copy.copy(bound.arguments)

    for arg_name, p in signature.parameters.items():
        if p.kind == inspect.Parameter.VAR_POSITIONAL:
            variable_args = arguments.pop(arg_name)
            argument_items = list(arguments.items())
            argument_items += [(arg_name, a) for a in variable_args]
            break
    else:
        argument_items = list(arguments.items())

    return_hint: typing.Type[T] = hints.pop("return", typing.Any)  # type: ignore

    mappings += [
        match_type(hints.get(name, typing.Any), arg)  # type: ignore
        for name, arg in argument_items
    ]
    try:
        matches: TypeVarMapping = merge_typevars(*mappings)
    except ValueError:
        raise TypeError(f"Couldn't merge mappings {mappings}")

    final_args = bound.args[1:] if is_classmethod else bound.args
    final_kwargs = bound.kwargs
    for arg in final_args:
        record_scoped_typevars(arg, *matches.keys())
    for kwarg in final_kwargs.values():
        record_scoped_typevars(kwarg, *matches.keys())
    return (final_args, final_kwargs, replace_typevars(matches, return_hint), matches)


def record_scoped_typevars(f: object, *additional_typevars: typing.TypeVar) -> None:  # type: ignore
    if not isinstance(f, types.FunctionType):
        return
    f.__scoped_typevars__ = frozenset(  # type: ignore
        {
            *get_all_typevars(get_function_type(f)),
            *get_typevars_in_scope(),
            *additional_typevars,
        }
    )


WrapperType = typing.Callable[
    [
        typing.Callable[..., T],
        typing.Tuple[object, ...],
        typing.Mapping[str, object],
        typing.Type[T],
    ],
    U,
]


_TYPEVARS_IN_SCOPE: typing.Counter[  # type: ignore
    typing.TypeVar
] = collections.Counter()


def get_typevars_in_scope() -> typing.Set[typing.TypeVar]:  # type: ignore
    return set(k for k, v in _TYPEVARS_IN_SCOPE.items() if v > 0)


@dataclasses.dataclass
class TypeVarScope:
    typevars_in_scope: typing.Tuple[typing.TypeVar, ...]  # type: ignore

    def __init__(self, *tvs: typing.TypeVar) -> None:  # type: ignore
        self.typevars_in_scope = tvs

    def __enter__(self) -> None:
        _TYPEVARS_IN_SCOPE.update(self.typevars_in_scope)

    def __exit__(self, *exc_details) -> None:
        _TYPEVARS_IN_SCOPE.subtract(self.typevars_in_scope)


@dataclasses.dataclass
class NewTypeVarScope:
    previous_typvars_in_scope: typing.Optional[  # type: ignore
        typing.Counter[typing.TypeVar]
    ] = None

    def __enter__(self) -> None:
        assert not self.previous_typvars_in_scope
        self.previous_typvars_in_scope = collections.Counter(_TYPEVARS_IN_SCOPE)
        _TYPEVARS_IN_SCOPE.clear()

    def __exit__(self, *exc_details) -> None:
        assert self.previous_typvars_in_scope
        _TYPEVARS_IN_SCOPE.update(self.previous_typvars_in_scope)


@dataclasses.dataclass(unsafe_hash=True)
class Infer(typing.Generic[T, U]):
    fn: typing.Callable[..., T]
    wrapper: WrapperType[T, U] = dataclasses.field(repr=False)

    def __post_init__(self):
        functools.update_wrapper(self, self.fn)

    def __call__(self, *args, **kwargs) -> U:
        *wrapper_args, typevars = infer_return_type(self.fn, None, False, args, kwargs)
        with TypeVarScope(*typevars.keys()):
            res = self.wrapper(self, *wrapper_args)  # type: ignore
        return res

    def __get__(self, instance, owner) -> BoundInfer[T, U]:
        is_classmethod = isinstance(self.fn, classmethod)
        is_property = isinstance(self.fn, property)
        if is_classmethod and is_property:
            raise NotImplementedError("classmethod properties are not supported")
        fn = self.fn
        if is_classmethod:
            fn = fn.__func__  # type: ignore
        if is_property:
            fn = fn.fget  # type: ignore
        if instance:
            method = BoundInfer(  # type: ignore
                fn, self.wrapper, get_type(instance), is_classmethod  # type: ignore
            )
            if is_property:
                return method(instance)  # type: ignore
            return functools.partial(method, instance)  # type: ignore
        return BoundInfer(  # type: ignore
            fn, self.wrapper, owner, is_classmethod  # type: ignore
        )

    def __repr__(self):
        return getattr(self.fn, "__name__", str(self.fn))


@dataclasses.dataclass(unsafe_hash=True)
class BoundInfer(typing.Generic[T, U]):
    fn: typing.Callable[..., T]
    wrapper: WrapperType[T, U] = dataclasses.field(repr=False)
    owner: typing.Type
    is_classmethod: bool

    def __post_init__(self):
        functools.update_wrapper(self, self.fn)

    def __call__(self, *args, **kwargs) -> U:
        *wrapper_args, typevars = infer_return_type(
            self.fn, self.owner, self.is_classmethod, args, kwargs
        )
        with TypeVarScope(*typevars.keys()):
            res = self.wrapper(self, *wrapper_args)  # type: ignore
        return res

    def __repr__(self):
        # Generic types are already formatted nicely
        owner_str = self.owner.__name__
        return f"{owner_str}.{self.fn.__name__}"


def infer(
    fn: typing.Callable[..., T], wrapper: WrapperType[T, U]
) -> typing.Callable[..., U]:
    """
    Wraps a function to return the args, kwargs, and the inferred return type based
    on the arguments.

    This raise a TypeError when called with types that are invalid.

    It refers on the explicit generic types of the arguments, it does not traverse
    them to check their types. That means if you pass in `[1, 2, 3]` it won't know
    this is a `typing.List[int]`. Instead it, will only know if you create a generic
    instance manually from a custom generic class like, `MyList[int](1, 2, 3)`.
    """

    return Infer(fn, wrapper)


def match_functions(
    fn_with_typevars: typing.Callable, fn: typing.Callable
) -> TypeVarMapping:
    if not isinstance(fn_with_typevars, BoundInfer) or not isinstance(fn, BoundInfer):
        if fn_with_typevars == fn:
            return {}
        raise TypeError(f"{fn_with_typevars} != {fn}")
    if fn_with_typevars.fn == fn.fn:
        return match_types(fn_with_typevars.owner, fn.owner)
    raise TypeError(f"{fn_with_typevars} != {fn}")


@dataclasses.dataclass(unsafe_hash=True)
class FunctionReplaceTyping:
    fn: typing.Callable
    typevars: HashableMapping[typing.TypeVar, typing.Type]  # type: ignore
    typevars_in_scope: typing.FrozenSet[typing.TypeVar]  # type: ignore
    inner_mapping: typing.Callable[[typing.Any], typing.Any]

    @classmethod
    def create(
        cls,
        fn: typing.Callable,
        typevars: TypeVarMapping,
        inner_mapping: typing.Callable[[typing.Any], typing.Any],
    ) -> typing.Callable:
        if isinstance(fn, FunctionReplaceTyping):
            return fn
        typevars_in_scope: typing.FrozenSet[TypeVar] = fn.__scoped_typevars__  # type: ignore
        if not typevars_in_scope:
            return fn
        if "fib_more" in str(fn):
            return fn

        typevars = HashableMapping(
            {k: v for k, v in typevars.items() if k in typevars_in_scope}
        )
        res = cls(fn, typevars, typevars_in_scope, inner_mapping)

        functools.update_wrapper(res, fn)
        res.__annotations__ = {
            k: replace_typevars(typevars, v)
            for k, v in typing_get_type_hints(fn).items()
        }

        return res

    def __call__(self, *args, **kwargs):
        with NewTypeVarScope():
            with TypeVarScope(*self.typevars_in_scope):
                return self.inner_mapping(self.fn(*args, **kwargs))  # type: ignore


@dataclasses.dataclass(unsafe_hash=True)
class Identity:
    def __call__(self, a):
        return a


def replace_fn_typevars(
    fn: T,
    typevars: TypeVarMapping,
    inner_mapping: typing.Callable[[T], T] = Identity(),
) -> T:
    if isinstance(fn, BoundInfer):
        return typing.cast(
            T,
            BoundInfer(  # type: ignore
                fn=fn.fn,
                wrapper=fn.wrapper,  # type: ignore
                is_classmethod=fn.is_classmethod,
                owner=replace_typevars(typevars, fn.owner),
            ),
        )
    if isinstance(fn, types.FunctionType):
        # Create new function by replacing typevars in existing function
        return FunctionReplaceTyping.create(fn, typevars, inner_mapping)  # type: ignore
    return fn


def get_fn_typevars(fn: object) -> TypeVarMapping:
    if isinstance(fn, BoundInfer):
        return match_types(get_origin_type(fn.owner), fn.owner)
    return {}



"""
Integeration with typez, to translate a "Rule" into a something that takes returns an iterable of
typez object we can use to display as we go.
"""

import dataclasses
import functools
import inspect
import types
import typing
import warnings

import metadsl
import typing_inspect
from typez import *

__all__ = ["ExpressionDisplay", "SHOW_MODULE"]


@functools.singledispatch
def metadsl_str(value: object) -> str:
    return str(value)


@metadsl_str.register
def metadsl_str_fn(fn: types.FunctionType) -> str:
    if not SHOW_TYPES or not hasattr(fn, "__scoped_typevars__"):
        return str(fn)
    return f"{fn} (scoped_typevars={fn.__scoped_typevars__})"  # type: ignore


@dataclasses.dataclass
class ExpressionDisplay:
    ref: metadsl.ExpressionReference
    typez_display: TypezDisplay = dataclasses.field(
        default_factory=lambda: TypezDisplay(Typez())
    )

    def __post_init__(self):
        ref = self.ref
        nodes = convert_to_nodes(ref)
        initial_node_id = str(ref.hash)
        self.typez_display.typez = Typez(
            states=States(initial=initial_node_id), nodes=nodes
        )

    def update(self, rule: str, label: typing.Optional[str] = None):
        new_nodes = convert_to_nodes(self.ref)
        old_nodes = self.typez_display.typez.nodes or []
        old_node_ids = set(node.id for node in old_nodes)
        self.typez_display.typez = dataclasses.replace(
            self.typez_display.typez,
            # Must keep in topo order with the root expression at the end
            nodes=(
                old_nodes + [node for node in new_nodes if node.id not in old_node_ids]
            ),
            states=dataclasses.replace(
                self.typez_display.typez.states,
                states=[
                    *(
                        self.typez_display.typez.states.states or []
                        if self.typez_display.typez.states
                        else []
                    ),
                    State(node=str(self.ref.hash), rule=rule, label=label),
                ],
            ),
        )

    def _ipython_display_(self):
        self.typez_display._ipython_display_()


def convert_to_nodes(ref: metadsl.ExpressionReference) -> Nodes:
    """
    Converts an expression into a node mapping.
    """
    nodes: Nodes = []
    for ref in ref.descendents:
        node: typing.Union[CallNode, PrimitiveNode]
        value = ref.expression
        if isinstance(value, metadsl.Expression):
            children = ref.children
            func_str = function_or_type_repr(value.function)
            node = CallNode(
                id=str(ref.hash),
                type_params=typevars_to_typeparams(
                    metadsl.typing_tools.get_fn_typevars(value.function)
                )
                or None,
                function=f"{func_str}: {value._type_str}" if SHOW_TYPES else func_str,
                args=[str(a) for a in children.args] or None,
                kwargs={k: str(v) for k, v in children.kwargs.items()} or None,
            )
        else:
            node = PrimitiveNode(
                id=str(ref.hash),
                type=function_or_type_repr(type(value)),
                repr=metadsl_str(value),
            )
        nodes.append(node)
    return nodes


def typevars_to_typeparams(
    typevars: metadsl.typing_tools.TypeVarMapping,
) -> typing.Dict[str, TypeInstance]:
    return {
        var.__name__: type_to_typeinstance(tp)  # type: ignore
        for var, tp in typevars.items()
    }


def type_to_typeinstance(tp: typing.Type) -> TypeInstance:
    """
    Converts a python type to the type instance, by folding through it
    """
    if issubclass(tp, metadsl.Expression):
        return DeclaredTypeInstance(
            type=function_or_type_repr(typing_inspect.get_origin(tp) or tp),
            params=typevars_to_typeparams(
                metadsl.typing_tools.match_types(
                    metadsl.typing_tools.get_origin_type(tp), tp
                )
            )
            or None,
        )
    return ExternalTypeInstance(repr=function_or_type_repr(tp))


_builtins = inspect.getmodule(int)


SHOW_MODULE = False

WARN_ON_TYPEVAR = False


def function_or_type_repr(o: typing.Union[typing.Type, typing.Callable]) -> str:
    """
    returns the name of the type or function prefixed by the module name if it isn't a builtin
    """
    if isinstance(o, typing.TypeVar):  # type: ignore
        if WARN_ON_TYPEVAR:
            warnings.warn(
                "Unresolved typevar in type argument, this means the typing is broken",
                RuntimeWarning,
            )
            return repr(o)
        else:
            raise RuntimeError("Unresolved typevar in type argument")

    # This is true for `typing.Any`
    if isinstance(o, typing._SpecialForm):
        return repr(o)

    name = o.__qualname__ or o.__name__
    # This is true for objects in typing
    if not name:
        return repr(o)
    if not SHOW_MODULE:
        return name
    module = inspect.getmodule(o)
    if module == _builtins:
        return name
    assert module
    return f"{module.__name__}.{name}"



import metadsl_core as mc
import metadsl_llvm as ml


__all__ = ["create_metadsl_fn"]

def create_metadsl_fn():
    # Constants
    int_type = ml.Type.create_int(32)
    zero = ml.Value.constant(int_type, 0)
    one = ml.Value.constant(int_type, 1)

    # Module reference
    mod_ref = ml.ModuleReference.create("fib")

    # Module builder
    mod_builder = ml.ModuleBuilder.create(mod_ref)

    # Function references

    mod_builder, fib_more_ref = ml.FunctionReference.create(
        mod_builder,
        ml.FunctionType.create(int_type, int_type, int_type, int_type),
        "fib_more",
        "fastcc",
    ).spread

    mod_builder, fib_ref = ml.FunctionReference.create(
        mod_builder, ml.FunctionType.create(int_type, int_type), "fib", "fastcc"
    ).spread

    # Function builders
    fib_more_builder = ml.FunctionBuilder.create(fib_more_ref)
    fib_builder = ml.FunctionBuilder.create(fib_ref)

    # Function args
    fib_n = fib_more_builder.arguments[mc.Integer.from_int(0)]

    fib_more_n = fib_builder.arguments[mc.Integer.from_int(0)]
    fib_more_a = fib_builder.arguments[mc.Integer.from_int(1)]
    fib_more_b = fib_builder.arguments[mc.Integer.from_int(2)]

    # Block references
    fib_more_builder, fib_more_entry_ref = ml.BlockReference.create(
        "entry", fib_more_builder
    ).spread
    fib_more_builder, fib_more_pred_cont_ref = ml.BlockReference.create(
        "pred_cont", fib_more_builder
    ).spread
    fib_more_builder, fib_more_not_pred_cont_ref = ml.BlockReference.create(
        "not_pred_cont", fib_more_builder
    ).spread
    fib_more_builder, fib_more_n_eq_one_ref = ml.BlockReference.create(
        "n_eq_one", fib_more_builder
    ).spread
    fib_more_builder, fib_more_n_neq_one_ref = ml.BlockReference.create(
        "n_neq_one", fib_more_builder
    ).spread

    fib_builder, fib_entry_ref = ml.BlockReference.create("entry", fib_builder).spread

    # Block builders

    fib_more_entry_builder = ml.Builder.create(fib_more_entry_ref)
    fib_more_pred_cont_builder = ml.Builder.create(fib_more_pred_cont_ref)
    fib_more_not_pred_cont_builder = ml.Builder.create(fib_more_not_pred_cont_ref)
    fib_more_n_eq_one_builder = ml.Builder.create(fib_more_n_eq_one_ref)
    fib_more_n_neq_one_builder = ml.Builder.create(fib_more_n_neq_one_ref)

    fib_entry_builder = ml.BlockBuilder.create(fib_entry_ref)

    # Values

    fib_entry_builder, fib_entry_value = fib_entry_builder.call(
        fib_more_ref, mc.Vec.create(fib_n, zero, one)
    ).spread
    fib_entry_builder = fib_entry_builder.ret(fib_entry_value)

    fib_more_entry_builder, pred_cont = fib_more_entry_builder.icmp_signed(
        ">", fib_more_n, one
    ).spread
    fib_more_entry_builder = fib_more_entry_builder.cbranch(
        pred_cont, fib_more_pred_cont_ref, fib_more_not_pred_cont_ref
    )

    fib_more_pred_cont_builder, minus1 = fib_more_pred_cont_builder.sub(
        fib_more_n, one
    ).spread
    fib_more_pred_cont_builder, ab = fib_more_pred_cont_builder.add(
        fib_more_a, fib_more_b
    )
    fib_more_pred_cont_builder, added = fib_more_pred_cont_builder.call(
        fib_more_ref, mc.Vec.create(minus1, fib_more_b, ab)
    ).spread
    fib_more_pred_cont_builder = fib_more_pred_cont_builder.ret(added)

    fib_more_not_pred_cont_builder, n_eq_1 = fib_more_not_pred_cont_builder.icmp_signed(
        "==", fib_more_n, one
    )
    fib_more_not_pred_cont_builder = fib_more_not_pred_cont_builder.cbranch(
        n_eq_1, fib_more_n_eq_one_ref, fib_more_n_neq_one_ref
    )

    fib_more_n_eq_one_builder = fib_more_n_eq_one_builder.ret(fib_more_b)

    fib_more_n_neq_one_builder = fib_more_n_neq_one_builder.ret(fib_more_a)

    # Blocks
    fib_more_entry_block = ml.Block.create(fib_more_entry_ref, fib_more_entry_builder)
    fib_more_pred_cont_block = ml.Block.create(
        fib_more_pred_cont_ref, fib_more_pred_cont_builder
    )
    fib_more_not_pred_cont_block = ml.Block.create(
        fib_more_not_pred_cont_ref, fib_more_not_pred_cont_builder
    )
    fib_more_n_eq_one_block = ml.Block.create(
        fib_more_n_eq_one_ref, fib_more_n_eq_one_builder
    )
    fib_more_n_neq_one_block = ml.Block.create(
        fib_more_n_neq_one_ref, fib_more_n_neq_one_builder
    )

    fib_entry_block = ml.Block.create(fib_entry_ref, fib_entry_builder)

    # Functions
    fib_more_fn = ml.Function.create(
        fib_more_ref,
        mc.Vec.create(
            fib_more_entry_block,
            fib_more_pred_cont_block,
            fib_more_not_pred_cont_block,
            fib_more_n_eq_one_block,
            fib_more_n_neq_one_block,
        ),
    )

    fib_fn = ml.Function.create(fib_ref, mc.Vec.create(fib_entry_block))

    # Module

    module = ml.Module.create(mod_ref, mc.Vec.create(fib_fn, fib_more_fn))

    c_int = ml.CType.c_int()
    c_func_type = ml.CFunctionType.create(c_int, c_int)

    return ml.compile_function(module, mod_builder, fib_ref, c_func_type)



from __future__ import annotations

import pytest
import typing

from .expressions import *
from .normalized import *
from .expressions_test import TEST_EXPRESSIONS


@pytest.mark.parametrize("expr", TEST_EXPRESSIONS)
def test_expression_reference_identity(expr):
    ref = ExpressionReference.from_expression(expr)
    assert ref.expression == expr


@expression
def a(e: typing.Any) -> typing.Any:
    ...


@expression
def b(e: typing.Any) -> typing.Any:
    ...


@expression
def c() -> typing.Any:
    ...


@expression
def d() -> typing.Any:
    ...


def test_replace():
    ref = ExpressionReference.from_expression(a(b(c())))

    assert tuple(child.expression for child in ref.descendents) == (
        c(),
        b(c()),
        a(b(c())),
    )

    ref.replace(a(b(d())))

    assert ref.expression == a(b(d()))

    assert tuple(child.expression for child in ref.descendents) == (
        d(),
        b(d()),
        a(b(d())),
    )


def test_replace_child():
    ref = ExpressionReference.from_expression(a(b(c())))

    assert tuple(child.expression for child in ref.descendents) == (
        c(),
        b(c()),
        a(b(c())),
    )

    child_ref = list(ref.descendents)[0]
    child_ref.replace(d())

    assert ref.expression == a(b(d()))

    assert tuple(child.expression for child in ref.descendents) == (
        d(),
        b(d()),
        a(b(d())),
    )


@expression
def e(a: typing.Any, b: typing.Any) -> typing.Any:
    ...


@expression
def f(e: typing.Any) -> typing.Any:
    ...


@expression
def g(e: typing.Any) -> typing.Any:
    ...


def test_graph_subtree():
    """
    Verify processed in correct order
    """
    orig = e(f(g(d())), d())
    ref = ExpressionReference.from_expression(orig)
    assert ref.expression == orig


def test_doesnt_remember_replacements():
    ref = ExpressionReference.from_expression(a(b(c())))
    ref.replace(a(b(d())))
    ref.replace(a(b(c())))

    assert ref.expression == a(b(c()))



import typing

import numpy
from metadsl import *

from . import numpy as numpy_api
from .abstraction import *
from .conversion import *
from .either import *
from .integer import *
from .maybe import *
from .pair import *
from .rules import *
from .vec import *

__all__ = ["ndarray_getitem"]

T = typing.TypeVar("T")


@expression
def unbox_integer(i: Integer) -> int:
    ...


@register_unbox  # type: ignore
@rule
def unbox_integer_rule(i: int) -> R[int]:
    return unbox_integer(Integer.from_int(i)), i


@expression
def arange(stop: int) -> numpy.ndarray:
    return numpy.arange(stop)


register_numpy_engine(default_rule(arange))


@register_unbox
@rule
def unbox_arange(i: Integer) -> R[numpy.ndarray]:
    return (numpy_api.arange_(i).to_ndarray(), arange(unbox_integer(i)))


@expression
def ndarray_getitem(
    self: numpy.ndarray, i: typing.Union[int, typing.Tuple[int, ...]]
) -> numpy.ndarray:
    # Always return array not scalars
    return numpy.array(self[i], ndmin=0)


register_numpy_engine(default_rule(ndarray_getitem))


@expression
def ndarray_add(self: numpy.ndarray, other: numpy.ndarray) -> numpy.ndarray:
    return self + other


register_numpy_engine(default_rule(ndarray_add))


@register_unbox  # type: ignore
@rule
def unbox_ndarray_add(a: numpy_api.NDArray, b: numpy_api.NDArray) -> R[numpy.ndarray]:
    return (a + b).to_ndarray(), ndarray_add(a.to_ndarray(), b.to_ndarray())


@expression
def unbox_idxs(
    idxs: Either[Integer, Vec[Integer]]
) -> typing.Union[int, typing.Tuple[int, ...]]:
    ...


@expression
def homo_tuple(*values: T) -> typing.Tuple[T, ...]:
    return tuple(values)


register(default_rule(homo_tuple))


@register_unbox  # type: ignore
@rule
def unbox_idxs_rule(
    i: Integer, ints: typing.Sequence[Integer]
) -> R[typing.Union[int, typing.Tuple[int, ...]]]:
    yield (unbox_idxs(Either[Integer, Vec[Integer]].left(i)), unbox_integer(i))
    yield (
        unbox_idxs(Either[Integer, Vec[Integer]].right(Vec.create(*ints))),
        lambda: homo_tuple(*(unbox_integer(i_) for i_ in ints)),  # type: ignore
    )


@register_unbox  # type: ignore
@rule
def unbox_ndarray_getitem(
    a: numpy_api.NDArray, idx: Either[Integer, Vec[Integer]]
) -> R[numpy.ndarray]:
    return a[idx].to_ndarray(), ndarray_getitem(a.to_ndarray(), unbox_idxs(idx))




from __future__ import annotations

import dataclasses
import itertools
import typing

from .typing_tools import *

__all__ = [
    "Expression",
    "expression",
    "ExpressionFolder",
    "ExpressionReplacer",
    "PlaceholderExpression",
    "IteratedPlaceholder",
    "create_iterated_placeholder",
]

T_expression = typing.TypeVar("T_expression", bound="Expression")


@dataclasses.dataclass(eq=False, repr=False)
class Expression(GenericCheck):
    """
    Top level object.
    Subclass this type and provide relevent methods for your type. Do not add any fields.

    Properties:

    Calling the function, after replacing the typevars in it (if it is a bound method),
    with the args and kwargs should resualt in an equivalent expression:

    replace_fn_typevars(self.function, self.typevars)(*self.args, **self.kwargs) == self
    
    The return type of the function, inferred by replacing the typevars in and with these args and kwargs,
    should match the type of the expression. If the return type of the function is not subclass of expression,
    then this should be a PlaceholderExpression of that type.
    """

    function: typing.Callable
    args: typing.List[object]
    kwargs: typing.Dict[str, object]

    def __str__(self):
        arg_strings = (str(arg) for arg in self.args)
        kwarg_strings = (f"{str(k)}={str(v)}" for k, v in self.kwargs.items())
        return f"{self._function_str}({', '.join(itertools.chain(arg_strings, kwarg_strings))})"

    @property
    def _function_str(self):
        return getattr(self.function, "__qualname__", str(self.function))

    @property
    def _type_str(self):
        t = get_type(self)
        return str(t)

    def __repr__(self):
        return (
            f"{self._type_str}({self.function}, {repr(self.args)}, {repr(self.kwargs)})"
        )

    def __eq__(self, value) -> bool:
        if not isinstance(value, Expression):
            return False

        return (
            self.function == value.function
            and self.args == value.args
            and self.kwargs == value.kwargs
        )


T = typing.TypeVar("T")


class PlaceholderExpression(Expression, OfType[T], typing.Generic[T]):
    """
    An expression that represents a type of `T`, for example T could be `int`.

    This is needed when a function returns a non expression type, it still has to return
    an expression under the covers until it has been replaced.

    It is also needed when using Wildcards in expressions when doing matching.
    """

    def __iter__(self):
        return iter((create_iterated_placeholder(self),))  # type: ignore


def extract_expression_type(t: typing.Type) -> typing.Type[Expression]:
    if issubclass(t, Expression):
        return t
    return PlaceholderExpression[t]  # type: ignore


T_callable = typing.TypeVar("T_callable", bound=typing.Callable)


def wrapper(fn, args, kwargs, return_type):
    expr_return_type = extract_expression_type(return_type)
    return expr_return_type(fn, list(args), kwargs)


def expression(fn: T_callable) -> T_callable:
    """
    Creates an expresion object by wrapping a Python function and providing a function
    that will take in the args and return an expression of the right type.
    """

    return typing.cast(T_callable, infer(fn, wrapper))


class IteratedPlaceholder(Expression, ExpandedType, typing.Generic[T]):
    pass


@expression
def create_iterated_placeholder(
    i: PlaceholderExpression[typing.Sequence[T]],
) -> IteratedPlaceholder[T]:
    """
    If a placeholder is of an iterable T, calling iter on it should return an iterable placeholder of the inner type.

    Used in matching with variable args.
    """
    ...


T_Expression = typing.TypeVar("T_Expression", bound="Expression")


@dataclasses.dataclass
class ExpressionFolder:
    """
    Traverses this expression graph and transforms each node, from the bottom up.
    """

    fn: typing.Callable[[object], object] = lambda e: e
    typevars: TypeVarMapping = dataclasses.field(default_factory=dict)

    def __call__(self, expr: object) -> object:
        fn: typing.Callable[[object], object] = self.fn  # type: ignore
        if isinstance(expr, Expression):
            return fn(
                (replace_fn_typevars(expr.function, self.typevars))(
                    *(self(arg) for arg in expr.args),
                    **{k: self(v) for k, v in expr.kwargs.items()},
                )
            )
        return fn(expr)


@dataclasses.dataclass
class ExpressionReplacer:
    """
    Replaces certain subexpression in an expression with values.
    """

    mapping: typing.Mapping
    folder: ExpressionFolder = dataclasses.field(init=False)

    def __post_init__(self):
        self.folder = ExpressionFolder(self.fn)

    def __call__(self, expr):
        return self.folder(expr)

    def fn(self, o):
        return self.mapping.get(o, o)



import metadsl as m
import metadsl_core as mc
import metadsl_llvm as ml


def test_fib():
    ##
    # Constants
    ##

    int_type = ml.Type.create_int(32)
    zero = ml.Value.constant(int_type, 0)
    one = ml.Value.constant(int_type, 1)

    ##
    # Module Ref
    ##
    mod_ref = ml.ModRef.create("fib")

    ##
    # Function Refs
    ##
    fib_more_ref = mod_ref.fn(
        "fib_more", ml.FnType.create(int_type, int_type, int_type, int_type), "fastcc",
    )

    fib_ref = mod_ref.fn("fib", ml.FnType.create(int_type, int_type), "fastcc")

    ##
    # Arguments
    ##

    fib_n = fib_ref.arguments[mc.Integer.from_int(0)]
    fib_more_n = fib_more_ref.arguments[mc.Integer.from_int(0)]
    fib_more_a = fib_more_ref.arguments[mc.Integer.from_int(1)]
    fib_more_b = fib_more_ref.arguments[mc.Integer.from_int(2)]

    ##
    # Block Refs
    ##
    fib_entry = fib_ref.block(True, "entry")
    fib_more_entry = fib_more_ref.block(True, "entry")
    fib_more_pred_cont = fib_more_ref.block(False, "pred_cont")
    fib_more_not_pred_cont = fib_more_ref.block(False, "not_pred_cont")
    fib_more_n_eq_one = fib_more_ref.block(False, "n_eq_one")
    fib_more_n_neq_one = fib_more_ref.block(False, "n_neq_one")

    ##
    # Values
    ##

    fib_entry_terminate = fib_entry.ret(
        fib_entry.call(fib_more_ref, mc.Vec.create(fib_n, zero, one))
    )

    fib_more_entry_terminate = fib_more_entry.cbranch(
        fib_more_entry.icmp_signed(">", fib_more_n, one),
        fib_more_pred_cont,
        fib_more_not_pred_cont,
    )

    fib_more_pred_cont_terminate = fib_more_pred_cont.ret(
        fib_more_pred_cont.call(
            fib_more_ref,
            mc.Vec.create(
                fib_more_pred_cont.sub(fib_more_n, one),
                fib_more_b,
                fib_more_pred_cont.add(fib_more_a, fib_more_b),
            ),
        )
    )

    fib_more_not_pred_cont_terminate = fib_more_not_pred_cont.cbranch(
        fib_more_not_pred_cont.icmp_signed("==", fib_more_n, one),
        fib_more_n_eq_one,
        fib_more_n_neq_one,
    )

    fib_more_n_eq_one_terminate = fib_more_n_eq_one.ret(fib_more_b)

    fib_more_n_neq_one_terminate = fib_more_n_neq_one.ret(fib_more_a)

    ##
    # Functions
    ##
    fib_fn = fib_ref.fn(fib_entry_terminate)
    fib_more_fn = fib_more_ref.fn(
        fib_more_entry_terminate,
        fib_more_pred_cont_terminate,
        fib_more_not_pred_cont_terminate,
        fib_more_n_eq_one_terminate,
        fib_more_n_neq_one_terminate,
    )

    ##
    # Module
    ##

    mod = mod_ref.mod(fib_fn, fib_more_fn)

    metadsl_fn = m.execute(ml.compile_function(mod, fib_ref))
    assert metadsl_fn(10) == 55


def test_add():
    ##
    # Constants
    ##

    int_type = ml.Type.create_int(32)
    one = ml.Value.constant(int_type, 1)

    mod_ref = ml.ModRef.create("add")
    fn_ref = mod_ref.fn(
        "add", ml.FnType.create(int_type, int_type, int_type), "fastcc",
    )

    l = fn_ref.arguments[mc.Integer.from_int(0)]
    r = fn_ref.arguments[mc.Integer.from_int(1)]

    block_ref = fn_ref.block(True)
    lr = block_ref.add(l, r)
    res = block_ref.add(lr, one)
    terminate = block_ref.ret(res)

    fn = fn_ref.fn(terminate)

    mod = mod_ref.mod(fn)
    real_fn = m.execute(ml.compile_function(mod, fn_ref))
    assert real_fn(10, 11) == 22



from __future__ import annotations

from metadsl import *

from .integer import *
from .pair import *


class TestPair:
    def test_left(self):
        assert execute(Pair.create(1, 2).left) == 1

    def test_right(self):
        assert execute(Pair.create(1, 2).right) == 2

    def test_left_type(self):
        assert execute(
            Pair.create(Integer.from_int(1), 2).left + Integer.from_int(1)
        ) == Integer.from_int(2)



from __future__ import annotations
import typing

from metadsl import *
from .rules import *
from .conversion import *
from .maybe import *


__all__ = ["Boolean"]

T = typing.TypeVar("T")


class Boolean(Expression):
    @expression
    @classmethod
    def create(cls, b: bool) -> Boolean:
        ...

    @expression
    def if_(self, true: T, false: T) -> T:
        ...


@register  # type: ignore
@rule
def if_(b: bool, l: T, r: T) -> R[T]:
    return Boolean.create(b).if_(l, r), lambda: l if b else r


@register_convert
@rule
def convert_bool(b: bool) -> R[Maybe[Boolean]]:
    """
    >>> execute(Converter[Boolean].convert(True), convert_bool) ==  Maybe.just(Boolean.create(True))
    True
    >>> list(convert_bool(ExpressionReference.from_expression(Converter[Boolean].convert("not bool"))))
    []
    """
    return Converter[Boolean].convert(b), Maybe.just(Boolean.create(b))



# """
# To create a version of an expression class that can
# deal with Pyhon values, you should subclass
# it and decorate each of the methods you want
# to wrap with `wrap_method`, supplying a new signature
# that takes in any python object and returns a wrapped
# expression.

# For a function, decorate it with `wrap`, passing in the
# original function first.
# """

# import typing
# import functools

# from .expressions import *
# from .conversion import *
# from .typing_tools import *

# __all__ = ["wrap", "wrap_method"]
# T = typing.TypeVar("T")


# T_callable = typing.TypeVar("T_callable", bound=typing.Callable)


# def wrap_method(fn: T_callable) -> T_callable:
#     """
#     Wraps a method of an expression to call it's super method but first converting
#     all arguments to expressions.
#     """
#     return_hint = typing.get_type_hints(fn)["return"]

#     def wrap_method_inner(self, *args, fn=fn, _return_type=return_hint) -> Expression:
#         original_fn = getattr(super(type(self), self), fn.__name__)
#         arg_hints = get_arg_hints(original_fn.__wrapped__, self)
#         converted_args = (
#             convert(hint, arg) if hint else arg for hint, arg in zip(arg_hints, args)
#         )
#         return original_fn(*converted_args, _return_type=_return_type)

#     return typing.cast(T_callable, wrap_method_inner)


# def wrap(original_fn: typing.Callable) -> typing.Callable[[T_callable], T_callable]:
#     def wrap_inner(fn: T_callable, original_fn=original_fn) -> T_callable:
#         """
#         Converts each of the arguments to type.
#         """
#         return_hint = typing.get_type_hints(fn)["return"]
#         arg_hints = get_arg_hints(original_fn.__wrapped__)

#         @functools.wraps(fn)
#         def wrap_inner_inner(
#             *args,
#             original_fn=original_fn,
#             arg_hints=arg_hints,
#             _return_type=return_hint
#         ) -> Expression:
#             converted_args = (
#                 convert(hint, arg) if hint else arg
#                 for hint, arg in zip(arg_hints, args)
#             )
#             return original_fn(*converted_args, _return_type=_return_type)

#         return typing.cast(T_callable, wrap_inner_inner)

#     return wrap_inner



"""
Meta language for describing DSLs in JSON.
"""
from __future__ import annotations

from dataclasses import dataclass
import dataclasses
from typing import *
import json
import jsonschema
import pathlib
import IPython.core.display

__all__ = [
    "Typez",
    "Definitions",
    "Kind",
    "Function",
    "Type",
    "TypeParameter",
    "DeclaredType",
    "ExternalType",
    "Nodes",
    "CallNode",
    "PrimitiveNode",
    "SHOW_TYPES",
    "TypeInstance",
    "DeclaredTypeInstance",
    "ExternalTypeInstance",
    "States",
    "State",
    "TypezDisplay",
]
__version__ = "0.3.0"

SHOW_TYPES = False

# with open(pathlib.Path(__file__).parent / "schema.json") as f:
#     typez_schema = json.load(f)

# All of these defintions are copied from
# `index.ts` and these two files should be manually
# kept synchronized


@dataclass(frozen=True)
class Typez:
    definitions: Optional[Definitions] = None
    nodes: Optional[Nodes] = None
    states: Optional[States] = None

    def asdict(self):
        """
        Turns this into a dict, removing keys with null values and
        validating it against the schema.
        """
        dict_ = dataclasses.asdict(
            self,
            dict_factory=lambda entries: {k: v for k, v in entries if v is not None},
        )
        # jsonschema.validate(dict_, typez_schema)
        return dict_

    def _repr_mimebundle_(self, include=None, exclude=None):
        return {"application/x.typez+json": self.asdict()}

    def __post_init__(self):
        """
        Validate that nodes are in topo order, with root at the end.
        """
        return
        if not self.nodes:
            return
        seen: Set[str] = set()
        for node in self.nodes:
            assert node.id not in seen
            seen.add(node.id)
            if not isinstance(node, CallNode):
                return
            for child_id in (node.args or []) + list((node.kwargs or {}).values()):
                assert child_id in seen


Definitions = Dict[str, Union["Kind", "Function"]]


@dataclass(frozen=True)
class Kind:
    params: Optional[List[str]] = None


@dataclass(frozen=True)
class Function:
    params: List[Tuple[str, Type]]
    return_: Type
    type_params: Optional[List[str]] = None
    rest_param: Optional[Tuple[str, Type]] = None


Type = Union["TypeParameter", "DeclaredType", "ExternalType"]


@dataclass(frozen=True)
class TypeParameter:
    param: str


@dataclass(frozen=True)
class DeclaredType:
    type: str
    params: Optional[Dict[str, Type]] = None


@dataclass(frozen=True)
class ExternalType:
    type: str
    repr: str


# TODO: Rename to expression
Nodes = List[Union["CallNode", "PrimitiveNode"]]


@dataclass(frozen=True)
class CallNode:
    id: str
    function: str
    type_params: Optional[Dict[str, TypeInstance]] = None
    args: Optional[List[str]] = None
    kwargs: Optional[Dict[str, str]] = None

    def __post_init__(self):
        """
        Make the args and kwargs hashable for easy hashing of the node
        to compute its id
        """
        # Use settattr because it is frozen
        if self.type_params is not None:
            object.__setattr__(self, "type_params", Hashabledict(self.type_params))
        if self.kwargs is not None:
            object.__setattr__(self, "kwargs", Hashabledict(self.kwargs))
        if self.args is not None:
            object.__setattr__(self, "args", Hashablelist(self.args))


@dataclass(frozen=True)
class PrimitiveNode:
    id: str
    type: str
    repr: str


TypeInstance = Union["DeclaredTypeInstance", "ExternalTypeInstance"]


@dataclass(frozen=True)
class DeclaredTypeInstance:
    type: str
    params: Optional[Dict[str, TypeInstance]] = None

    def __post_init__(self):
        if self.params is not None:
            object.__setattr__(self, "params", Hashabledict(self.params))


@dataclass(frozen=True)
class ExternalTypeInstance:
    repr: str


@dataclass(frozen=True)
class States:
    initial: str
    states: Optional[List["State"]] = None


@dataclass(frozen=True)
class State:
    node: str
    rule: str
    label: Optional[str] = None


@dataclass
class TypezDisplay:
    """
    A display object for typez objects. If you set the `typez` property after calling `display` it will update
    the existing display.
    """

    typez: Typez
    _typez: Typez = dataclasses.field(init=False, repr=False)
    _handle: Optional[IPython.core.display.DisplayHandle] = dataclasses.field(
        default=None, init=False, repr=False
    )

    def _ipython_display_(self):
        self._handle = IPython.core.display.display(self.typez, display_id=True)

    @property  # type: ignore
    def typez(self):
        return self._typez

    @typez.setter
    def typez(self, value: Typez):
        self._typez = value
        if self._handle:
            self._handle.update(self.typez)


class Hashabledict(dict):
    """
    Dict that hashses to its key, value pairs.

    https://stackoverflow.com/a/16162138/907060
    """

    def __hash__(self):
        return hash(frozenset(self.items()))


class Hashablelist(list):
    def __hash__(self):
        return hash(frozenset(self))



"""
Core data types and replacements for metadsl
"""

from .abstraction import *
from .conversion import *
from .either import *
from .maybe import *
from .rules import *
from .boolean import *
from .pair import *
from .vec import *
from .integer import *
from .numpy import *
from .numpy_engine import *
from .function import *

__version__ = "0.3.0"



from metadsl import execute, ExpressionReference
from .conversion import *
from .maybe import *
from .rules import *


class TestConvertIdentity:
    def test_matches_type(self):
        assert execute(Converter[int].convert(1)) == Maybe.just(1)

    def test_doesnt_match_type(self):
        assert not list(
            convert_identity_rule(
                ExpressionReference.from_expression(Converter[str].convert(1))
            )
        )

    def test_matches_convert(self):
        assert execute(Converter[int].convert(Maybe.just(1))) == Maybe.just(1)
        assert (
            execute(Converter[int].convert(Maybe[int].nothing()))
            == Maybe[int].nothing()
        )


class TestConvertToMaybe:
    def test_just(self):
        assert execute(Converter[Maybe[int]].convert(1)) == Maybe.just(Maybe.just(1))

    def test_nothing(self):
        assert execute(Converter[Maybe[int]].convert(None)) == Maybe.just(
            Maybe[int].nothing()
        )



from __future__ import annotations

import numpy
from metadsl import *

from .vec import *
from .integer import *
from .conversion import *
from .rules import *
from .maybe import *
from .abstraction import *
from .pair import *
from .either import *


__all__ = ["arange", "IndxType", "NDArray"]


class IntCompat(Expression):
    """
    API should be work like Python's `int` type.
    """

    @expression
    def __add__(self, other: object) -> IntCompat:
        return IntCompat.from_maybe_integer(
            (Converter[Integer].convert(self) & Converter[Integer].convert(other)).map(
                Abstraction[Pair[Integer, Integer], Integer].from_fn(
                    lambda p: p.left + p.right
                )
            )
        )

    @expression
    @classmethod
    def from_maybe_integer(cls, i: Maybe[Integer]) -> IntCompat:
        ...


register_convert(default_rule(IntCompat.__add__))


@register_convert
@rule
def convert_to_integer(i: Maybe[Integer]) -> R[Maybe[Integer]]:
    return Converter[Integer].convert(IntCompat.from_maybe_integer(i)), i


class TupleIntCompat(Expression):
    @expression
    def __getitem__(self, i: object) -> IntCompat:
        return IntCompat.from_maybe_integer(
            (Converter[Vec[Integer]].convert(self) & Converter[Integer].convert(i)).map(
                Abstraction[Pair[Vec[Integer], Integer], Integer].from_fn(
                    lambda p: p.left[p.right]
                )
            )
        )

    @expression
    @classmethod
    def from_vec_integer(cls, i: Maybe[Vec[Integer]]) -> TupleIntCompat:
        ...


register_convert(default_rule(TupleIntCompat.__getitem__))


@register_convert
@rule
def convert_to_vec_integer(i: Maybe[Vec[Integer]]) -> R[Maybe[Vec[Integer]]]:
    return Converter[Vec[Integer]].convert(TupleIntCompat.from_vec_integer(i)), i


IndxType = Either[Integer, Vec[Integer]]


class NDArray(Expression):
    @expression
    def __getitem__(self, idxs: IndxType) -> NDArray:
        ...

    @expression
    def __add__(self, other: NDArray) -> NDArray:
        ...

    @expression
    def to_ndarray(self) -> numpy.ndarray:
        ...

    @expression
    @classmethod
    def from_ndarray(self, n: numpy.ndarray) -> NDArray:
        ...


@register_unbox
@rule
def box_unbox_ndarray(n: numpy.ndarray) -> R[numpy.ndarray]:
    return NDArray.from_ndarray(n).to_ndarray(), n


class NDArrayCompat(Expression):
    @expression
    def __getitem__(self, idxs: object) -> NDArrayCompat:
        return NDArrayCompat.from_ndarray(
            (
                Converter[NDArray].convert(self)
                & Converter[Either[Integer, Vec[Integer]]].convert(idxs)
            ).map(
                Abstraction[
                    Pair[NDArray, Either[Integer, Vec[Integer]]], NDArray
                ].from_fn(lambda p: p.left[p.right])
            )
        )

    @expression
    def __add__(self, other: NDArrayCompat) -> NDArrayCompat:
        ...

    @expression
    @classmethod
    def from_ndarray(cls, n: Maybe[NDArray]) -> NDArrayCompat:
        ...

    @expression
    def to_ndarray(self) -> numpy.ndarray:
        ...


@register_unbox
@rule
def box_unbox_ndarray_compat(n: NDArray) -> R[numpy.ndarray]:
    return (NDArrayCompat.from_ndarray(Maybe.just(n)).to_ndarray(), n.to_ndarray())


@register_convert
@rule
def add_compat(l: NDArray, r: NDArray) -> R[NDArrayCompat]:
    return (
        NDArrayCompat.from_ndarray(Maybe.just(l))
        + NDArrayCompat.from_ndarray(Maybe.just(r)),
        NDArrayCompat.from_ndarray(Maybe.just(l + r)),
    )


register_convert(default_rule(NDArrayCompat.__getitem__))


@register_convert
@rule
def convert_to_ndarray(i: Maybe[NDArray]) -> R[Maybe[NDArray]]:
    return Converter[NDArray].convert(NDArrayCompat.from_ndarray(i)), i


@expression
def arange(stop: object) -> NDArrayCompat:
    return NDArrayCompat.from_ndarray(
        Converter[Integer].convert(stop).map(Abstraction.from_fn(arange_))
    )


register_convert(default_rule(arange))


@expression
def arange_(stop: Integer) -> NDArray:
    ...



from __future__ import annotations

import typing

import pytest

from .typing_tools import *

T = typing.TypeVar("T")


def i(fn):
    return infer(fn, lambda *args: args)


def test_simple():
    @i
    def _simple_return() -> int:
        ...

    assert _simple_return() == (_simple_return, (), {}, int)


def test_generic():
    @i
    def _generic_return(arg: T) -> T:
        ...

    assert _generic_return(123.0) == (_generic_return, (123.0,), {}, float)


class _GenericClassMethod(GenericCheck, typing.Generic[T]):
    @i
    def method(self) -> T:
        ...


def test_generic_class_arg():
    c = _GenericClassMethod[int]()
    assert c.method() == (_GenericClassMethod[int].method, (c,), {}, int)


def test_generic_arg_call_on_class():
    c = _GenericClassMethod[int]()
    assert _GenericClassMethod.method(c) == (_GenericClassMethod.method, (c,), {}, int)
    assert _GenericClassMethod[int].method(c) == (
        _GenericClassMethod[int].method,
        (c,),
        {},
        int,
    )

    c2 = _GenericClassMethod[_GenericClassMethod[int]]()
    assert c2.method() == (
        _GenericClassMethod[_GenericClassMethod[int]].method,
        (c2,),
        {},
        _GenericClassMethod[int],
    )


def test_generic_arg_call_on_class_generic_class():
    c: _GenericClassMethod = _GenericClassMethod()
    assert _GenericClassMethod[int].method(c) == (
        _GenericClassMethod[int].method,
        (c,),
        {},
        int,
    )


class _GenericClassCreate(GenericCheck, typing.Generic[T]):
    @i
    @classmethod
    def create(cls) -> T:
        ...

    @i
    @classmethod
    def create_item(cls, item: T) -> T:
        ...


def test_create():
    assert _GenericClassCreate[int].create() == (
        _GenericClassCreate[int].create,
        (),
        {},
        int,
    )

    assert _GenericClassCreate[_GenericClassMethod[int]].create() == (
        _GenericClassCreate[_GenericClassMethod[int]].create,
        (),
        {},
        _GenericClassMethod[int],
    )


def test_create_type_from_arg():

    assert _GenericClassCreate.create_item(123) == (
        _GenericClassCreate.create_item,
        (123,),
        {},
        int,
    )

    assert _GenericClassCreate[int].create_item(123) == (
        _GenericClassCreate[int].create_item,
        (123,),
        {},
        int,
    )


class _NonGenericClass:
    @i
    def method(self) -> int:
        ...


def test_non_generic_method():

    c = _NonGenericClass()
    assert c.method() == (_NonGenericClass.method, (c,), {}, int)
    assert _NonGenericClass.method(c) == (_NonGenericClass.method, (c,), {}, int)


def test_variable_args():
    @i
    def many_args(*x: T) -> T:
        ...

    assert many_args(1) == (many_args, (1,), {}, int)
    with pytest.raises(TypeError):
        many_args(1, "hj")


def test_variable_args_empty():
    @i
    def many_args(*x: int) -> int:
        ...

    assert many_args() == (many_args, (), {}, int)


def test_keyword_args():
    @i
    def many_args(a: int, b: T) -> T:
        ...

    assert many_args(b="df", a=123) == (many_args, (123, "df"), {}, str)


def test_keyword_default():
    @i
    def default_kwarg(b: T = 10) -> T:  # type: ignore
        ...

    assert default_kwarg() == (default_kwarg, (10,), {}, int)
    assert default_kwarg(b="df") == (default_kwarg, ("df",), {}, str)


def test_tuple_for_sequence():
    @i
    def fn(xs: typing.Sequence[T]) -> T:  # type: ignore
        ...

    assert fn((1, 2)) == (fn, ((1, 2),), {}, int)


def test_fn_args():
    @i
    def fn(f: typing.Callable) -> int:
        ...

    f = lambda i: 10
    assert fn(f) == (fn, (f,), {}, int)


def test_fn_args_generic():
    def f(i: int) -> str:
        ...

    assert _GenericClassCreate[typing.Callable[[int], str]].create_item(f) == (
        _GenericClassCreate[typing.Callable[[int], str]].create_item,
        (f,),
        {},
        typing.Callable[[int], str],
    )


U = typing.TypeVar("U")


class Both(typing.Generic[T, U]):
    ...


class ConflictingTypevars(typing.Generic[T]):
    @i
    def __add__(self, other: ConflictingTypevars[U]) -> ConflictingTypevars[Both[T, U]]:
        ...


def test_generic_class_conflicting_typevars():
    """
    We should make sure that if you use a typevar in a generic class, it gets resolved locally.
    """
    l = ConflictingTypevars[int]()
    r = ConflictingTypevars[str]()
    assert l + r == (
        ConflictingTypevars[int].__add__,
        (l, r),
        {},
        ConflictingTypevars[Both[int, str]],
    )


V = typing.TypeVar("V")


class ConflictingTypevarsMultiple(typing.Generic[T, U]):
    @i
    def __add__(
        self, other: ConflictingTypevarsMultiple[U, V]
    ) -> ConflictingTypevarsMultiple[T, V]:
        ...


def test_generic_class_conflicting_multiple_typevars():
    l = ConflictingTypevarsMultiple[int, str]()
    r = ConflictingTypevarsMultiple[str, float]()
    assert l + r == (
        ConflictingTypevarsMultiple[int, str].__add__,
        (l, r),
        {},
        ConflictingTypevarsMultiple[int, float],
    )


class C(typing.Generic[T]):
    @i
    def __add__(self, other: T) -> T:
        ...

    @i
    @classmethod
    def create(cls, arg: T) -> C[T]:
        ...


class TestGetType:
    def test_bound_infer_method(self):

        assert get_type(C[int]().__add__) == typing.Callable[[int], int]

    def test_bound_infer_method_on_class(self):

        assert get_type(C[int].__add__) == typing.Callable[[C[int], int], int]

    def test_bound_infer_classmethod(self):

        assert get_type(C[int].create) == typing.Callable[[int], C[int]]



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['metadsl']

package_data = \
{'': ['*']}

install_requires = \
['typing_extensions', 'typing_inspect', 'python-igraph>=0.8.0']

extras_require = \
{'dev': ['jupyterlab>=1.0.0', 'nbconvert', 'pudb', 'beni'],
 'doc': ['sphinx',
         'sphinx-autodoc-typehints',
         'sphinx_rtd_theme',
         'recommonmark',
         'nbsphinx',
         'ipykernel',
         'IPython',
         'sphinx-autobuild'],
 'test': ['pytest>=3.6.0',
          'pytest-cov',
          'pytest-mypy',
          'pytest-randomly',
          'pytest-xdist',
          'pytest-pudb',
          'mypy']}

setup(name='metadsl',
      version='0.3.0',
      description='Library to help create DSLs in Python.',
      author='Saul Shanabrook',
      author_email='s.shanabrook@gmail.com',
      url='https://github.com/Quansight-Labs/metadsl',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.8',
     )



from __future__ import annotations

import typing
import typing_inspect
import pytest

from .expressions import *

T = typing.TypeVar("T")


@expression
def fn(a: T, b: T) -> T:
    ...


@expression
def value_fn(a: int) -> Expression:
    ...


class Subclass(Expression):
    pass


@expression
def subclass_fn(a: int) -> Subclass:
    ...


def test_expression_subclass() -> None:
    assert isinstance(subclass_fn(1), Subclass)


class Generic(Expression, typing.Generic[T]):
    def get_inner_type(self) -> typing.Type[T]:
        return typing_inspect.get_args(typing_inspect.get_generic_type(self))[0]

    @expression
    def get(self) -> T:
        ...

    @expression
    @classmethod
    def create(cls) -> Generic[T]:
        ...


@expression
def instance_arg_fn(a: int) -> Generic[int]:
    ...


def test_expression_subclass_generic() -> None:
    assert instance_arg_fn(10).get_inner_type() == int


@expression
def mutable_fn(a: typing.List) -> Expression:
    ...


class SubclassWithMethod(Expression):
    @expression
    def __add__(self, other: SubclassWithMethod) -> SubclassWithMethod:
        ...


def test_classmethod():
    c = Generic[Subclass].create()
    assert isinstance(c.get(), Subclass)


@expression
def create_method_subclass(i: int) -> SubclassWithMethod:
    ...


@expression
def variadic(*args: int) -> Expression:
    ...


def test_variadic():
    assert variadic(1, 2).args == [1, 2]


@expression
def create_variadic_ints() -> typing.Sequence[int]:
    ...


def test_iterated():
    assert variadic(1, 2, *create_variadic_ints(), 3, 4) == Expression(
        variadic,
        [
            1,
            2,
            IteratedPlaceholder[int](
                create_iterated_placeholder,
                [
                    PlaceholderExpression[typing.Sequence[int]](
                        create_variadic_ints, [], {}
                    )
                ],
                {},
            ),
            3,
            4,
        ],
        {},
    )


TEST_EXPRESSIONS: typing.Iterable[object] = [
    value_fn(123),
    fn(value_fn(10), value_fn(11)),
    fn(subclass_fn(1), subclass_fn(2)),
    instance_arg_fn(10),
    create_method_subclass(10) + create_method_subclass(10),
    mutable_fn([1, 2, 3]),  # mutable value that doesn't have hash
    variadic(1, 2, 3),
    Generic[int].create(),
    Generic[Generic[int]].create(),
]


@pytest.mark.parametrize("instance", TEST_EXPRESSIONS)
def test_fold_identity(instance) -> None:
    """
    You should be able to decompose any instances into it's type and expression and recompose it.
    """
    result = ExpressionFolder(lambda e: e)(instance)
    assert instance == result


U = typing.TypeVar("U")


def test_fold_typevars_identity():
    assert ExpressionFolder()(Generic[T].create()) == Generic[T].create()
    assert ExpressionFolder()(Generic[U].create()) == Generic[U].create()


def test_fold_typevars_replace():
    assert (
        ExpressionFolder(typevars={T: int})(Generic[T].create())
        == Generic[int].create()
    )
    assert (
        ExpressionFolder(typevars={U: int})(Generic[U].create())
        == Generic[int].create()
    )


def test_fold_typevars_replace_skip_missing():
    assert (
        ExpressionFolder(typevars={U: int})(Generic[T].create()) == Generic[T].create()
    )
    assert (
        ExpressionFolder(typevars={T: int})(Generic[U].create()) == Generic[U].create()
    )



#!/usr/bin/env python
# coding: utf-8

# In[1]:


from metadsl import *
from metadsl_core import *

import metadsl_visualize

import typez
typez.SHOW_TYPES = True


# In[2]:


@FunctionTwo.from_fn
def add(a: Integer, b: Integer) -> Integer:
    return a + b


# In[3]:


res = add(Integer.from_int(1), Integer.from_int(1))


# In[5]:


res._ipython_display_()


# In[ ]:







"""
Pattern matching allows you to define replacement rules by creating a template that should resemble
an expression, with certain leaf nodes as Wildcards. These can match any node in the target expression.

The values at these wildcards are captured and passed to another function to compute a resulting expression
to use as the replacement.

Conceptually, this is similar to regular expression matching / finite state transducers. Technically,
this is a form of symbolic tree transducers, but not implemented in a mathematically consistant way. If it was,
then we could more easily combine matches to make them execute faster and compute properties about them. This would
be good to add at a later date, and could be done without having users change their code.
"""

import typing
import dataclasses
import functools
import types
import inspect
import typing_inspect
import functools

from .expressions import *
from .dict_tools import *
from .normalized import *
from .typing_tools import *
from .rules import Rule, Replacement


__all__ = ["R", "NoMatch", "rule", "default_rule", "create_wildcard"]

T = typing.TypeVar("T", bound=Expression)

R_single = typing.Tuple[T, typing.Union[typing.Callable[[], T], T]]
R = typing.Union[typing.Generator[R_single[T], None, None], R_single[T]]
# Mapping from wildcards to the matching pattern and replacement
MatchFunctionType = typing.Callable[..., R[T]]


class NoMatch(Exception):
    pass


def rule(fn: MatchFunctionType) -> Rule:
    """
    Creates a new rule given a callable that accepts wildcards and returns
    the match value and the replacement value.
    """

    return MatchRule(fn)


T_Callable = typing.TypeVar("T_Callable", bound=typing.Callable)


def default_rule(fn: typing.Callable) -> Rule:
    """
    Creates a rule based on the body of a passed in expression function
    """
    return DefaultRule(fn)


@dataclasses.dataclass
class DefaultRule:
    fn: typing.Callable
    inner_fn: typing.Callable = dataclasses.field(init=False, repr=False)

    def __str__(self):
        return f"{self.inner_fn.__module__}.{self.inner_fn.__qualname__}"

    def __post_init__(self):
        self.inner_fn = self.fn.__wrapped__  # type: ignore

    def __call__(self, ref: ExpressionReference) -> typing.Iterable[Replacement]:
        """
        This rule should match whenever the expression is this function.

        Then, all it has to do is get the type variable mapping given the current args
        and apply it to the return value. This is so that if the body uses generic type
        variables, they are turned into the actual instantiations. 
        """
        expr = ref.expression
        if not isinstance(expr, Expression):
            return

        fn = self.fn

        args = expr.args

        # If any of the args are placeholders, don't match!
        if any(
            isinstance(arg, PlaceholderExpression)
            for arg in args + list(expr.kwargs.values())
        ):
            return None

        typevars: TypeVarMapping = infer_return_type(
            expr.function.fn,  # type: ignore
            getattr(expr.function, "owner", None),
            getattr(expr.function, "is_classmethod", False),
            tuple(args),
            expr.kwargs,
        )[-1]
        if isinstance(fn, BoundInfer) and isinstance(expr.function, BoundInfer):
            if fn.fn != expr.function.fn:
                return None
            if fn.is_classmethod:
                args = [
                    typing.cast(object, replace_typevars(typevars, fn.owner))
                ] + args

        elif fn != expr.function:
            return None
        with TypeVarScope(*typevars.keys()):
            new_expr = self.inner_fn(*args, **expr.kwargs)
            result = ReplaceTypevarsExpression(typevars)(new_expr)
        ref.replace(result)
        yield Replacement(str(self))


class Wildcard(Expression, typing.Generic[T]):
    """
    Insert this somewhere in an expression tree to represent any expression or value of type T.
    """


@expression
def create_wildcard_expession(t: typing.Type[T], obj: object) -> T:
    """
    The object inside is only needed for identity checks, to different
    two wildcards if the objects are different.
    """
    ...


def create_wildcard(t: typing.Type[T]) -> T:
    """
    Returns a wildcard of type "t"
    """
    return create_wildcard_expession(t, object())


# a number of (wildcard expression, replacement) pairs
WildcardMapping = typing.Mapping[Expression, object]


@dataclasses.dataclass
class MatchRule:
    """
    Creates a replacement rule given a function that maps from wildcard inputs
    to two things, a template expression tree and a replacement thunk.

    If the template matches an expression, it will be replaced with the result of the thunk, replacing
    the input args with the nodes at their locations in the template.

    You can also return None from the rule to signal that it won't match.
    """

    matchfunction: MatchFunctionType

    # the wildcards that are present in the template
    wildcards: typing.List[Expression] = dataclasses.field(init=False)

    results: typing.List[R] = dataclasses.field(init=False)

    def __str__(self):
        return f"{self.matchfunction.__module__}.{self.matchfunction.__qualname__}"

    def __post_init__(self):
        functools.update_wrapper(self, self.matchfunction)
        # Create one wildcard` per argument
        self.wildcards = [create_wildcard(a) for a in get_arg_hints(self.matchfunction)]

        # Call the function first to create a template with the wildcards
        result = self.matchfunction(*self.wildcards)
        self.results = (
            list(result)  # type: ignore
            if inspect.isgeneratorfunction(self.matchfunction)
            else [result]
        )

    def __call__(self, ref: ExpressionReference) -> typing.Iterable[Replacement]:
        expr = ref.expression
        for i, result in enumerate(self.results):
            template, _ = result
            try:
                typevars, wildcards_to_nodes = match_expression(  # type: ignore
                    self.wildcards, template, expr
                )
            except NoMatch:
                continue

            args = [
                wildcards_to_nodes.get(wildcard, wildcard)
                for wildcard in self.wildcards
            ]
            _, expression_thunk = (
                list(self.matchfunction(*args))[i]
                if inspect.isgeneratorfunction(self.matchfunction)
                else self.matchfunction(*args)
            )
            try:
                result_expr: object = (
                    expression_thunk()
                    if isinstance(expression_thunk, types.FunctionType)
                    else expression_thunk
                )
            except NoMatch:
                continue
            with TypeVarScope(*typevars.keys()):
                result_expr = ReplaceTypevarsExpression(typevars)(result_expr)
                ref.replace(result_expr)
            yield Replacement(str(self))
            return


@dataclasses.dataclass(frozen=True)
class ReplaceTypevarsExpression:
    """
    Use a class instead of a function so we can partially apply it and have equality based on typevars 
    """

    typevars: TypeVarMapping

    def __call__(self, expression: object) -> object:
        """
        Replaces all typevars found in the classmethods of an expression.
        """
        typevars = self.typevars
        if isinstance(expression, Expression):
            new_args = [self(a) for a in expression.args]
            new_kwargs = {k: self(v) for k, v in expression.kwargs.items()}
            new_fn = replace_fn_typevars(expression.function, typevars)
            return replace_typevars(
                typevars, typing_inspect.get_generic_type(expression)
            )(new_fn, new_args, new_kwargs)
        return replace_fn_typevars(
            expression,
            typevars,
            ReplaceTypevarsExpression(typevars=HashableMapping(typevars)),
        )


def match_expression(
    wildcards: typing.List[Expression], template: object, expr: object
) -> typing.Tuple[TypeVarMapping, WildcardMapping]:
    """
    Returns a mapping of wildcards to the objects at that level, or None if it does not match.

    template: the expression with wildcards in it
    expr: the expression without wildcards to match against

    A wildcard can match either an expression or a value. If it matches two nodes, they must be equal.
    """
    if template in wildcards:
        # If we are matching against a placeholder and the expression is not resolved to that placeholder, don't match.
        if (
            isinstance(template, PlaceholderExpression)
            and isinstance(expr, Expression)
            and not typing_inspect.is_typevar(
                typing_inspect.get_args(typing_inspect.get_generic_type(template))[0]
            )
        ):
            raise NoMatch

        # Match type of wildcard with type of expression
        try:
            return (
                match_values(template, expr),
                UnhashableMapping(Item(typing.cast(Expression, template), expr)),
            )
        except TypeError:
            raise NoMatch

    if isinstance(expr, Expression):
        if not isinstance(template, Expression):
            raise NoMatch
        # Any typevars in the template that are unbound should be matched with their
        # versions in the expr

        try:
            fn_type_mapping: TypeVarMapping = match_functions(
                template.function, expr.function
            )
        except TypeError:
            raise NoMatch

        if set(expr.kwargs.keys()) != set(template.kwargs.keys()):
            raise TypeError("Wrong kwargs in match")

        template_args: typing.Iterable[object]
        expr_args: typing.Iterable[object]

        # Process args in the template that can represent any number of args.
        # These are the "IteratedPlaceholder"s
        # Allow one iterated placeholder in the template args
        # For example fn(a, b, [...], *c, d, e, [...])
        # Here `c` should take as many args as it can between the ends,
        # Each of those should be matched against the inner
        iterated_args = [
            arg for arg in template.args if isinstance(arg, IteratedPlaceholder)
        ]
        if iterated_args:
            # Only support one iterated arg for now
            # TODO: Support more than one, would require branching
            assert len(iterated_args) == 1
            # template args, minus the iterators, is the minimum length of the values
            # If they have less values than this, raise an error
            min_n_args = len(template.args) - 1
            if len(expr.args) < min_n_args:
                raise NoMatch("Wrong number of args in match")
            template_args_ = list(template.args)
            (template_iterated,) = iterated_args
            template_index_iterated = list(template.args).index(template_iterated)

            # Swap template iterated with inner wildcard
            (template_args_[template_index_iterated],) = template_iterated.args
            template_args = template_args_

            expr_args = collapse_tuple(
                tuple(expr.args),
                template_index_iterated,
                # The number we should preserve on the right, is the number of template
                # args after index
                len(template.args) - template_index_iterated - 1,
            )

        else:
            if len(template.args) != len(expr.args):
                raise NoMatch("Wrong number of args in match")
            template_args = template.args
            expr_args = expr.args

        type_mappings, expr_mappings = list(
            zip(
                *(
                    match_expression(wildcards, arg_template, arg_value)
                    for arg_template, arg_value in zip(template_args, expr_args)
                ),
                *(
                    match_expression(wildcards, template.kwargs[key], expr.kwargs[key])
                    for key in template.kwargs.keys()
                ),
            )
        ) or ((), ())
        try:
            merged_typevars: TypeVarMapping = merge_typevars(
                fn_type_mapping, *type_mappings
            )
        except TypeError:
            raise NoMatch
        try:
            return (
                merged_typevars,
                safe_merge(*expr_mappings, dict_constructor=UnhashableMapping),
            )
        except ValueError:
            raise NoMatch
    if template != expr:
        raise NoMatch
    return match_values(template, expr), UnhashableMapping()


def collapse_tuple(t: typing.Tuple, l: int, r: int) -> typing.Tuple:
    """
    Collapses the middle of a tuple into another nested tuple, keeping l on the left side
    uncollapsed and r on the right side un collapsed

    >>> collapse_tuple(tuple(range(10)), 3, 4)
    (0, 1, 2, (3, 4, 5), 6, 7, 8, 9)
    >>> collapse_tuple(tuple(range(5)), 3, 0)
    (0, 1, 2, (3, 4))
    """
    if r == 0:
        return tuple([*t[:l], t[l:]])
    return tuple([*t[:l], t[l:-r], *t[-r:]])



# import typing

# Import visualize so uses custom execution if in ipython
from . import visualize


def pytest_assertrepr_compare(op, left, right):
    pass
    # import graphviz
    # from .visualize import visualize
    # from metadsl import Expression

    # if op == "==" and (isinstance(left, Expression) or isinstance(right, Expression)):
    #     d = graphviz.Digraph()
    #     seen: typing.Set[int] = set()
    #     visualize(left, d, seen)
    #     visualize(right, d, seen)
    #     d.render(cleanup=True, view=True)

    #     return [f"{left!r} != {right!r}"]



print("Importing")
from metadsl import *
from metadsl_core import *

print("Finished imports")


one = Integer.from_int(1)
zero = Integer.from_int(0)


@FunctionThree.from_fn_recursive
def fib_more(
    fn: FunctionThree[Integer, Integer, Integer, Integer],
    n: Integer,
    a: Integer,
    b: Integer,
) -> Integer:
    pred_cont = n > one
    minus1 = n - one
    ab = a + b
    added = fn(minus1, b, ab)

    n_eq_1 = n.eq(one)
    return pred_cont.if_(added, n_eq_1.if_(b, a))


@FunctionOne.from_fn
def fib(n: Integer) -> Integer:
    return fib_more(n, zero, one)


print("Starting execution")
print(execute(fib(Integer.from_int(10))))
